// RungeKutta:
// Uses the Fourth-Order Runge-Kutta method to integrate the equations
// of motion of a satellite on a transfer from Earth to Mars (type
// dydt = f(t,y)), taking into consideration the gravity of the sun,
// Earth, and Mars for all time t

// Input: 
// Initial time -- InitTime
// Final time -- FinalTime
// Time step -- Step
// Initial heliocentric state -- InitWrtSun

// Output:
// Final heliocentric state -- SatWrtSun
// Final Mars-relative state -- SatWrtMars

// Thomas Reppert 03/17/07

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ArrayMath.h"
#include "AstroConstants.h"
#include "MathConstants.h"

// Declare the offset angle calculator
double CalcOffset();

// Declare the planet-relative state calculators
void SatVsEarth(int, double*, double*);
void SatVsMars(int, double, double*, double*);

// Declare the PlanetVsSun state calculators
void EarthVsSun(int, double*);
void MarsVsSun(int, double, double*);

// Declare the derivative calculator for each slope estimate
void Derivative(int, double, double*, double*);

void RungeKutta(int InitTime, int FinalTime, int NominalStep,
								double* InitWrtSun, double* SatWrtSun, double* SatWrtMars)
{
	// Declare slope estimates k_i, averaged slope estimate phi
  double k1[7], k2[7], k3[7], k4[7], phi[7];
	// Declare scaled slope estimates k_sc_i and scaled average phi_sc
	double k1_sc[7], k2_sc[7], k3_sc[7], phi_sc[7];
	// Declare the temporary and old heliocentric state place holders
  double StateTemp[7], SatWrtSunOld[7];
	// Declare the Earth-centric state and planets' SOI radii
	double SatWrtEarth[7], rE_SOI, rM_SOI;
	// Declare the position magnitudes wrt the planets and Mars offset
	double PosMagWrtEarth, PosMagWrtMars, Offset;
	// Declare the iterated time and variable time step
	int Time, VariableStep;
	// Declare file opening error flag
	//errno_t err;
	// Declare pointers to files for writing
	//FILE *FSatWrtSun;
	//FILE *FSatWrtEarth;
	//FILE *FSatWrtMars;

	// Define switches for enabling write to output files
	//#define SAT_WRT_SUN
	//#define SAT_WRT_EARTH
	//#define SAT_WRT_MARS

	#ifdef SAT_WRT_SUN
		// Check to make sure the output file opens correctly
		// If it doesn't open, leave the program
		if(err = fopen_s(&FSatWrtSun, "SatWrtSun02.txt", "w") != 0)
		{ 
			printf("*** Output file could not be opened.\n");
			exit(EXIT_FAILURE);
		}
	#endif

	#ifdef SAT_WRT_EARTH
		// Check to make sure the output file opens correctly
		// If it doesn't open, leave the program
		if(err = fopen_s(&FSatWrtEarth, "SatWrtEarth02.txt", "w") != 0)
		{ 
			printf("*** Output file could not be opened.\n");
			exit(EXIT_FAILURE);
		}
	#endif

	#ifdef SAT_WRT_MARS
		// Check to make sure the output file opens correctly
		// If it doesn't open, leave the program
		if(err = fopen_s(&FSatWrtMars, "SatWrtMars02.txt", "w") != 0)
		{ 
			printf("*** Output file could not be opened.\n");
			exit(EXIT_FAILURE);
		}
	#endif

	// Calculate the offset angle between Earth and Mars
	Offset = CalcOffset();

	// Calculate the initial planet-relative state vectors
	SatVsEarth(InitTime,InitWrtSun, SatWrtEarth);
	SatVsMars(InitTime,Offset,InitWrtSun, SatWrtMars);

	// Initialize the current time and state vector for integration
	Time = InitTime;
	Equal(7, SatWrtSun, InitWrtSun);

	// Write the initial conditions to the output files
	#ifdef SAT_WRT_SUN
		fprintf(FSatWrtSun,"%10i %12.7e %12.7e %12.7e\n",
			InitTime,SatWrtSun[1],SatWrtSun[2],SatWrtSun[3]);
	#endif
	#ifdef SAT_WRT_EARTH
		fprintf(FSatWrtEarth,"%10i %12.7e %12.7e %12.7e\n",
			InitTime,SatWrtEarth[1],SatWrtEarth[2],SatWrtEarth[3]);
	#endif
	#ifdef SAT_WRT_MARS
		fprintf(FSatWrtMars,"%10i %12.7e %12.7e %12.7e %12.7e %12.7e\n",
			InitTime,SatWrtMars[1],SatWrtMars[2],SatWrtMars[3],SatWrtMars[4],SatWrtMars[5]);
	#endif

	// Use the planets' SOI radii to operate the variable time step
	rE_SOI = 916600.0; // Earth's SOI radius, km
	rM_SOI = 577400.0; // Mars's SOI radius, km

	// Loop from the initial time to final time, with time step VariableStep (s)
	// NOTE: Make sure VariableStep is divisible by 2 (integer division
	// when using the Derivative function)
	while (Time < FinalTime)
	{
		// Use a variable time step to speed up the integration
		PosMagWrtEarth = sqrt(SatWrtEarth[1]*SatWrtEarth[1] +
			SatWrtEarth[2]*SatWrtEarth[2] + SatWrtEarth[3]*SatWrtEarth[3]);
		PosMagWrtMars = sqrt(SatWrtMars[1]*SatWrtMars[1] +
			SatWrtMars[2]*SatWrtMars[2] + SatWrtMars[3]*SatWrtMars[3]);

		// Increase the time step when the satellite's position wrt both
		// Earth and Mars is greater than 1.5*r_SOI of both planets
		if ((PosMagWrtEarth > 1.5*rE_SOI) && (PosMagWrtMars > 1.5*rM_SOI))
			VariableStep = NominalStep*1000;
    else VariableStep = NominalStep;

		// Estimate slope at Time and assign to k1
    Derivative(Time,Offset,SatWrtSun, k1);

		// Scale the k1 slope estimate by h/2
    ScaleArray(7,VariableStep/2.0,k1, k1_sc);

		// Assign (y + k1*h/2) value to temporary state
    Add2Arrays(7,SatWrtSun,k1_sc, StateTemp);

		// Use the new StateTemp to estimate slope at (Time + h/2) as k2
    Derivative(Time + VariableStep/2,Offset,StateTemp, k2);

		// Scale the k2 slope estimate by h/2
		ScaleArray(7,VariableStep/2.0,k2, k2_sc);
        
		// Assign (y + k2*h/2) value to temporary state
		Add2Arrays(7,SatWrtSun,k2_sc, StateTemp);
        
		// Use the new StateTemp to estimate slope at (Time + h/2) as k3
		Derivative(Time + VariableStep/2,Offset,StateTemp, k3);
		
		// Scale the k3 slope estimate by h
		ScaleArray(7,VariableStep,k3, k3_sc);
		
		// Assign (y + k3*h) value to temporary state
		Add2Arrays(7,SatWrtSun,k3_sc, StateTemp);
		
		// Use the new StateTemp to estimate slope at (Time + h) as k4
		Derivative(Time + VariableStep,Offset,StateTemp, k4);
		
		// Scale the k2 slope estimate to twice its original value
		ScaleArray(7,2,k2, k2_sc);
		
		// Scale the k3 slope estimate to twice its original value
		ScaleArray(7,2,k3, k3_sc);

		// Calculate the averaged slope estimate phi
		// Note: phi weighted as (k1 + 2*k2 + 2*k3 + k4)/6
		Add4Arrays(7,k1,k2_sc,k3_sc,k4, phi_sc);
		ScaleArray(7,VariableStep/6.0,phi_sc, phi); 
		
		// Assign current state to SatWrtSunOld in preparation for calculating
		// the new state
		Equal(7, SatWrtSunOld, SatWrtSun);
		
		// Add the old state to quantity phi*h for the new helio state
		Add2Arrays(7,SatWrtSunOld,phi, SatWrtSun);

		// Increment the time variable
		Time = Time + VariableStep;

		// Compute the new satellite planet-relative states
		SatVsEarth(Time,SatWrtSun, SatWrtEarth);
		SatVsMars(Time,Offset,SatWrtSun, SatWrtMars);
		
		// Print the values of the propagated state vectors
		#ifdef SAT_WRT_SUN
			fprintf(FSatWrtSun,"%10i %12.7e %12.7e %12.7e\n",
				Time,SatWrtSun[1],SatWrtSun[2],SatWrtSun[3]);
		#endif
		#ifdef SAT_WRT_EARTH
			fprintf(FSatWrtEarth,"%10i %12.7e %12.7e %12.7e\n",
				Time,SatWrtEarth[1],SatWrtEarth[2],SatWrtEarth[3]);
		#endif
		#ifdef SAT_WRT_MARS
			fprintf(FSatWrtMars,"%10i %12.7e %12.7e %12.7e %12.7e %12.7e\n",
				Time,SatWrtMars[1],SatWrtMars[2],SatWrtMars[3],SatWrtMars[4],SatWrtMars[5]);
		#endif
  }

	#ifdef SAT_WRT_SUN
		fclose(FSatWrtSun);
	#endif
	#ifdef SAT_WRT_EARTH
		fclose(FSatWrtEarth);
	#endif
	#ifdef SAT_WRT_MARS
		fclose(FSatWrtMars);
	#endif

  return;
}



// Derivative:
// Calculates the value of each slope estimate k for the Fourth-Order
// Runge-Kutta integration

// Input:
// Current time -- Time
// Initial offset angle between Earth and Mars -- Offset
// Satellite's heliocentric state -- SatWrtSun

// Output:
// Derivative vector -- Deriv

// Thomas Reppert 03/17/07

void Derivative(int Time, double Offset, double* SatWrtSun,
								double* Deriv)
{
	// Declare the planet-relative state vectors
	double SatWrtEarth[7], SatWrtMars[7];
	// Declare the body-relative position magnitudes
	double PosMagWrtSun, PosMagWrtEarth, PosMagWrtMars;

	// Calculate the planet-relative states
	SatVsEarth(Time,SatWrtSun, SatWrtEarth);
	SatVsMars(Time,Offset,SatWrtSun, SatWrtMars);

	// Calculate the body-relative position magnitudes
	PosMagWrtSun = sqrt(SatWrtSun[1]*SatWrtSun[1] +
		SatWrtSun[2]*SatWrtSun[2] + SatWrtSun[3]*SatWrtSun[3]);
	PosMagWrtEarth = sqrt(SatWrtEarth[1]*SatWrtEarth[1] +
		SatWrtEarth[2]*SatWrtEarth[2] + SatWrtEarth[3]*SatWrtEarth[3]);
	PosMagWrtMars = sqrt(SatWrtMars[1]*SatWrtMars[1] +
		SatWrtMars[2]*SatWrtMars[2] + SatWrtMars[3]*SatWrtMars[3]);

	// Calculate the slope estimate components
	// ---------------------------------------
	Deriv[1] = SatWrtSun[4];
	Deriv[2] = SatWrtSun[5];
	Deriv[3] = SatWrtSun[6];
	Deriv[4] = -MU_SUN/(PosMagWrtSun*PosMagWrtSun*PosMagWrtSun)*SatWrtSun[1] -
		MU_EARTH/(PosMagWrtEarth*PosMagWrtEarth*PosMagWrtEarth)*SatWrtEarth[1] -
		MU_MARS/(PosMagWrtMars*PosMagWrtMars*PosMagWrtMars)*SatWrtMars[1];
	Deriv[5] = -MU_SUN/(PosMagWrtSun*PosMagWrtSun*PosMagWrtSun)*SatWrtSun[2] -
		MU_EARTH/(PosMagWrtEarth*PosMagWrtEarth*PosMagWrtEarth)*SatWrtEarth[2] -
		MU_MARS/(PosMagWrtMars*PosMagWrtMars*PosMagWrtMars)*SatWrtMars[2];
	Deriv[6] = -MU_SUN/(PosMagWrtSun*PosMagWrtSun*PosMagWrtSun)*SatWrtSun[3] -
		MU_EARTH/(PosMagWrtEarth*PosMagWrtEarth*PosMagWrtEarth)*SatWrtEarth[3] -
		MU_MARS/(PosMagWrtMars*PosMagWrtMars*PosMagWrtMars)*SatWrtMars[3];

	return;
}



// SatVsEarth:
// Computes the satellite's Earth-relative state at the specified
// time t during the transfer orbit

// Input: 
// Current time -- t
// Current satellite heliocentric state -- SatHelio

// Output:
// Satellite's Earth-relative state -- SatWrtEarth

// Thomas Reppert 03/17/07

void SatVsEarth(int t, double* SatHelio, double* SatWrtEarth)
{
	// Declare Earth's heliocentric state vector
	double EarthHelio[7];
	
	// Compute Earth's heliocentric state vector at the current time t
	EarthVsSun(t, EarthHelio); 

	// Satellite's position wrt Earth, km
	// ----------------------------------
	SatWrtEarth[1] = SatHelio[1] - EarthHelio[1]; // Position component x
	SatWrtEarth[2] = SatHelio[2] - EarthHelio[2]; // Position component y
	SatWrtEarth[3] = SatHelio[3] - EarthHelio[3]; // Position component z

	// Satellite's velocity wrt Earth, km/s
	// ------------------------------------
	SatWrtEarth[4] = SatHelio[4] - EarthHelio[4]; // Velocity component Vx
	SatWrtEarth[5] = SatHelio[5] - EarthHelio[5]; // Velocity component Vy
	SatWrtEarth[6] = SatHelio[6] - EarthHelio[6]; // Velocity component Vz

	return;
}



// SatVsMars:
// Computes the satellite's Mars-relative state at the specified
// time t during the transfer orbit

// Input:
// Current time -- t
// Initial offset angle between Earth and Mars -- offset
// Current satellite heliocentric state -- SatHelio

// Output:
// Satellite's Mars-relative state -- SatWrtMars

// Thomas Reppert 03/17/07

void SatVsMars(int t, double offset, double* SatHelio,
							 double* SatWrtMars)
{
	// Declare Mars' heliocentric state vector
	double MarsHelio[7];

	// Compute Mars' heliocentric state vector at the current time t
	MarsVsSun(t,offset, MarsHelio);

	// Satellite's position wrt Mars, km
	// ---------------------------------
	SatWrtMars[1] = SatHelio[1] - MarsHelio[1]; // Position component x
	SatWrtMars[2] = SatHelio[2] - MarsHelio[2]; // Position component y
	SatWrtMars[3] = SatHelio[3] - MarsHelio[3]; // Position component z
	
	// Satellite's velocity wrt Mars, km/s
	// -----------------------------------
	SatWrtMars[4] = SatHelio[4] - MarsHelio[4]; // Velocity component Vx
	SatWrtMars[5] = SatHelio[5] - MarsHelio[5]; // Velocity component Vy
	SatWrtMars[6] = SatHelio[6] - MarsHelio[6]; // Velocity component Vz

	return;
}



// EarthVsSun:
// Calculates Earth's heliocentric state vector at the specified time t
// for a circular orbit about the sun

// Input:
// Current time -- t

// Output:
// Earth's heliocentric state vector -- EarthHelio

// Thomas Reppert 03/16/07

void EarthVsSun(int t, double* EarthHelio)
{
	// Initialize Earth's mean orbit rate, rad/s
	double n_earth = sqrt(MU_SUN/(SMA_EARTH*SMA_EARTH*SMA_EARTH));

	// Earth's current heliocentric position vector, km
	// ------------------------------------------------
	EarthHelio[1] = SMA_EARTH*cos(n_earth*t);
	EarthHelio[2] = SMA_EARTH*sin(n_earth*t);
	EarthHelio[3] =										   0.0;

	// Earth's current heliocentric velocity vector, km/s
	// --------------------------------------------------
	EarthHelio[4] = -SMA_EARTH*n_earth*sin(n_earth*t);
	EarthHelio[5] =  SMA_EARTH*n_earth*cos(n_earth*t);
	EarthHelio[6] =														    0.0;

	return;
}



// MarsVsSun:
// Calculates Mars' heliocentric state vector at the specified time t
// for a circular orbit about the sun

// Input:
// Current time -- t
// Initial offset angle between Earth and Mars -- offset

// Output:
// Mars' heliocentric state vector -- MarsHelio

// Thomas Reppert 03/16/07

void MarsVsSun(int t, double offset, double* MarsHelio)
{
	// Initialize Mars' mean orbit rate, rad/s
	double n_mars = sqrt(MU_SUN/(SMA_MARS*SMA_MARS*SMA_MARS));

	// Mars' current heliocentric position vector, km
	// ----------------------------------------------
	MarsHelio[1] = SMA_MARS*cos(n_mars*t + offset);
	MarsHelio[2] = SMA_MARS*sin(n_mars*t + offset);
	MarsHelio[3] =														 0.0;

	// Mars' current heliocentric velocity vector, km/s
	// ------------------------------------------------
	MarsHelio[4] = -SMA_MARS*n_mars*sin(n_mars*t + offset);
	MarsHelio[5] =  SMA_MARS*n_mars*cos(n_mars*t + offset);
	MarsHelio[6] =														         0.0;

	return;
}



// CalcOffset:
// Calculates the orbital offset angle between Earth and Mars

// Output:
// Offset angle Offset, rad

// Thomas Reppert 03/12/07

double CalcOffset()
{
	double n_mars, a, P, Offset;

	// Mars's mean orbit rate, rad/s
	n_mars = sqrt(MU_SUN/(SMA_MARS*SMA_MARS*SMA_MARS));
	// Semi-major axis of the Hohmann transfer, km
	a = (SMA_EARTH + SMA_MARS)/2;
	// Period of the Hohmann transfer, s
	P = PI*sqrt((a*a*a)/MU_SUN);
	// Offset angle between Earth and Mars, rad
	Offset = PI - n_mars*P;

	return Offset;
}
