// Sensitivity:
// Uses sensitivities to solve a two-point boundary value problem
// for the initial state that yields the desired final state

// Input:
// Initial time -- InitTime
// Final time -- FinalTime
// Time step -- Step
// Initial heliocentric state vector -- InitWrtSun

// Output:
// Corrected initial heliocentric state vector -- InitWrtSun
// Final heliocentric state vector -- FinalWrtSun
// Final Mars-relative state vector -- FinalWrtMars

// Thomas Reppert 03/30/07

#include     <stdio.h>
#include    <stdlib.h>
#include      <math.h>
#include "ArrayMath.h"

// Declare the LU decomposition function
void LU(int, double**, double*, double*);

// Declare the Runge-Kutta integrator
void RungeKutta(int, int, int, double*, double*, double*);

void Sensitivity(int InitTime, int FinalTime, int Step,
								 double* InitWrtSun, double* FinalWrtSun, double* FinalWrtMars)
{
	// Initialize the loop counter, ST counter, and ST rank
	int k = 1, j = 0, Rank = 2;
	// Initialize the perturbation value
	double Eps = 1.e-7;
	// Declare the perturbed initial heliocentric state vectors
	double IHCSPrimeVx[7], IHCSPrimeVy[7];
	// Declare the perturbed final heliocentric state vectors
	double FHCSPrimeVx[7], FHCSPrimeVy[7];
	// Declare the perturbed final Mars-relative state vectors
	double FMCSPrimeVx[7], FMCSPrimeVy[7];
	// Declare the Mars-relative altitudes and radial velocity components
	double FinalMCAlt, FinalMCVr, FMCAPrimeVx, FMCVrPrimeVx,
		FMCAPrimeVy, FMCVrPrimeVy;
	// Declare the error in the final state and the IC adjustment vector
	double Error[3], Adjust[3];
	// Declare the state transition matrix
	double** Phi;

	// Prepare the state transition matrix
	Phi = (double**) calloc(Rank+1, sizeof(double*));
	for (j = 0; j <= Rank; j++) Phi[j] = (double*) 
		calloc(Rank+1, sizeof(double));

	// Loop until within the desired tolerance
	while(1)
	{
		// Call the Runge-Kutta integrator
		RungeKutta(InitTime,FinalTime,Step,InitWrtSun, FinalWrtSun,
			FinalWrtMars);
		FinalMCAlt = sqrt(FinalWrtMars[1]*FinalWrtMars[1] + FinalWrtMars[2]*
			FinalWrtMars[2]);
		FinalMCVr = FinalWrtMars[1]/FinalMCAlt*FinalWrtMars[4] +
			sin(acos(FinalWrtMars[1]/FinalMCAlt))*FinalWrtMars[5];

		// Calculate the errors in the final state
		// Desired final state at an altitude of 600 km above Mars' surface
		// without any radial Mars-relative velocity component
		Error[1] = FinalMCAlt - 4000.0;
		Error[2] = FinalMCVr - 0.0;

		// If the errors are within tolerance, or counter is 20, then break
		if (fabs(Error[1]) < 1.e0 && fabs(Error[2]) < 1.e-1) break;
		else if (k == 20) {
			printf("Sensitivity algorithm DNC for tf = %i s.\n",FinalTime);
			exit(EXIT_FAILURE);}

		// Otherwise, use sensitivities to adjust the initial conditions
		else
		{
			// Prepare the initial condition perturbations
			Equal(7, IHCSPrimeVx, InitWrtSun);
			Equal(7, IHCSPrimeVy, InitWrtSun);
			IHCSPrimeVx[4] = InitWrtSun[4] + Eps;
			IHCSPrimeVy[5] = InitWrtSun[5] + Eps;

			// Perturb Vx and recalculate the final state vector
			RungeKutta(InitTime,FinalTime,Step,IHCSPrimeVx, FHCSPrimeVx,
				FMCSPrimeVx);
			FMCAPrimeVx = sqrt(FMCSPrimeVx[1]*FMCSPrimeVx[1] + FMCSPrimeVx[2]*
				FMCSPrimeVx[2]);
			FMCVrPrimeVx = FMCSPrimeVx[1]/FMCAPrimeVx*FMCSPrimeVx[4] +
				sin(acos(FMCSPrimeVx[1]/FMCAPrimeVx))*FMCSPrimeVx[5];

			// Perturb Vy and recalculate the final state vector
			RungeKutta(InitTime,FinalTime,Step,IHCSPrimeVy, FHCSPrimeVy,
				FMCSPrimeVy);
			FMCAPrimeVy = sqrt(FMCSPrimeVy[1]*FMCSPrimeVy[1] + FMCSPrimeVy[2]*
				FMCSPrimeVy[2]);
			FMCVrPrimeVy = FMCSPrimeVy[1]/FMCAPrimeVy*FMCSPrimeVy[4] +
				sin(acos(FMCSPrimeVy[1]/FMCAPrimeVy))*FMCSPrimeVy[5];

			// Calculate the first column of the state transition matrix
			Phi[1][1] = (FMCAPrimeVx - FinalMCAlt)/Eps;
			Phi[2][1] = (FMCVrPrimeVx - FinalMCVr)/Eps;

			// Calculate the second column of the state transition matrix
			Phi[1][2] = (FMCAPrimeVy - FinalMCAlt)/Eps;
			Phi[2][2] = (FMCVrPrimeVy - FinalMCVr)/Eps;

			// Use the ST matrix inverse to adjust the initial conditions
			LU(Rank,Phi,Error, Adjust);
			InitWrtSun[4] = InitWrtSun[4] - Adjust[1];
			InitWrtSun[5] = InitWrtSun[5] - Adjust[2];
		}

		// Increment the loop counter
		k = k + 1;
	}

	// Deallocate the memory used for the ST matrix
	for (j = 0; j <= Rank; j++) free(Phi[j]);
	free(Phi);

	return;
}
